import sys
from .lennybot import LennyBot
import os
import argparse


def _version()->str:
    if "__version__" in locals():
        return locals()["__version__"]
    if "__version__" in globals():
        return globals()["__version__"]
    if os.path.exists("version.txt"):
        with open("version.txt") as file_ptr:
            return file_ptr.read()
    return "-"


def _find_config(args):
    if args.config is not None:
        for item in args.config:
            return item
    if "LB_CONFIG_FILE" in os.environ.keys():
        return os.environ["LB_CONFIG_FILE"]
    if os.path.exists("config.yaml"):
        return "config.yaml"
    if os.path.exists("config.yml"):
        return "config.yml"
    raise Exception(
        "Configuration file not found, create 'config.yaml' or set LB_CONFIG_FILE to point to the config file")


def _arg_parser():
    parser = argparse.ArgumentParser(
        prog="lennybot",
        description="""
            The Lennybot checks for updates, creates a plan to update resources and can also apply this plan.
        """)
    parser.add_argument('action', metavar='action', type=str, nargs='?',
                        choices=['ci', 'plan', 'apply'], default='ci',
                        help='The action which should be executed. Has to be one of "plan", "apply" or "ci". Default is "ci"')
    parser.add_argument('-p', '--plan', dest='plan', type=str, required=False,
                        help='The filename of plan which should be saved or loaded')
    parser.add_argument('-c', '--config', dest='config', type=str, required=False, action='append',
                        help='A config value in the format key.subkey=value')
    parser.add_argument('-v', '--version',
                        action='version', version=_version())
    return parser.parse_args()


def main() -> int:
    args = _arg_parser()
    config_file = _find_config(args)
    app = LennyBot(config_file)
    # Execute plan only
    if args.action == "plan":
        plan = app.plan()
        app.save_plan("lennybot.plan", plan)
    # Execute plan and apply
    elif args.action == "apply" and args.plan is None:
        plan = app.plan()
        app.apply(plan)
    # Execute apply with given plan
    elif args.action == "apply" and args.plan is not None:
        plan = app.load_plan(args.plan)
        app.apply(plan)
    elif args.action == "ci":
        app.ci_setup()
        plan = app.plan()
        result = app.apply(plan)
        app.ci_finalize(plan, result)
    else:
        print("Unexpected Arguments")
        return 1
    return 0



from .source_github import GithubSource
from .source_github_query import GithubQuerySource
from .isource import ISource
from ...config import LennyBotSourceConfig

def create_source(name, config: LennyBotSourceConfig, github) -> ISource:
    source_type = config.type
    if source_type == "github":
        return GithubSource(name, config, github)
    if source_type == "github-query":
        return GithubQuerySource(name, config, github)
    raise Exception(f"Unknown Source Type: {source_type}")


from typing import List
import logging

from ..model.plan import LennyBotPlan

class ApplyService:

    def __init__(self) -> None:
        pass

    def apply(self, plan: LennyBotPlan):
        state=plan.state
        if not state.is_valid():
            raise Exception("Invalid State")
        for action in plan.actions:
            try:
                action.run()
                state.update_version(action.application, action.target_version)
            except Exception as exception:
                print(f"Exception during action execution for {action.application}")
                raise exception
        state.save()


from typing import List
import yaml
import os
import logging

CONFIGURATION_OPTIONS = {
    "github": {
        "type": "object",
        "properties": {
            "token": {
                "type": "string",
                "required": True,
                "attribute": "_github_token"
            },
            "pr": {
                "type": "object",
                "attribute": "_github_pr",
                "properties": {
                    "enabled": {
                        "type": "bool",
                        "attribute": "_enabled"
                    },
                    "repository": {
                        "type": "bool",
                        "attribute": "_repository"
                    },
                    "branchPrefix": {
                        "type": "bool",
                        "attribute": "_branch_prefix"
                    },
                }
            }
        }
    },
    "state": {
        "type": "object",
        "properties": {
            "file": {
                "type": "string",
                "required": True,
                "attribute": "_state_file"
            }
        }
    },
    "applications": {
        "type": "list",
        "required": True,
        "class": "LennyBotAppConfig",
        "attribute": "_applications",
        "properties": {
            "name": {
                "type": "string",
                "required": True,
                "attribute": "_name"
            },
            "source": {
                "type": "object",
                "attribute": "_source",
                "properties": {
                    "type": {
                        "type": "string",
                        "required": True,
                        "attribute": "_type"
                    },
                    "repository": {
                        "type": "string",
                        "required": True,
                        "attribute": "_repository"
                    },
                    "regex": {
                        "type": "string",
                        "attribute": "_regex"
                    },
                }
            },
            "actions": {
                "type": "list",
                "class": "LennyBotActionConfig",
                "attribute": "_actions",
                "properties": {
                    "type": {
                        "type": "string",
                        "attribute": "_type"
                    },
                    "image": {
                        "type": "string",
                        "attribute": "_image"
                    },
                    "kustomizePath": {
                        "type": "string",
                        "attribute": "_kustomize_path"
                    },
                    "tagPattern": {
                        "type": "string",
                        "attribute": "_tag_pattern"
                    },
                    "url": {
                        "type": "string",
                        "attribute": "_url"
                    },
                    "target": {
                        "type": "string",
                        "attribute": "_target"
                    },
                    "targetFile": {
                        "type": "string",
                        "attribute": "_target_file"
                    },
                    "yamlPath": {
                        "type": "string",
                        "attribute": "_yaml_path"
                    },
                    "valuePattern": {
                        "type": "string",
                        "attribute": "_value_pattern"
                    }
                }
            }
        }
    }
}


class LennyBotSourceConfig:

    def __init__(self) -> None:
        self._type = None
        self._repository = None
        self._regex = None

    @property
    def type(self) -> str:
        return self._type

    @property
    def repository(self) -> str:
        return self._repository

    @property
    def regex(self) -> str:
        return self._regex


class LennyBotActionConfig:

    def __init__(self) -> None:
        self._type = None
        self._image = None
        self._kustomize_path = None
        self._tag_pattern = None
        self._target = None
        self._url = None
        self._target_file = None
        self._yaml_path = None
        self._value_pattern = None

    @property
    def type(self) -> str:
        return self._type

    @property
    def image(self) -> str:
        return self._image

    @property
    def kustomize_path(self) -> str:
        return self._kustomize_path

    @property
    def tag_pattern(self) -> str:
        return self._tag_pattern

    @property
    def target(self) -> str:
        return self._target

    @property
    def url(self) -> str:
        return self._url

    @property
    def target_file(self) -> str:
        return self._target_file

    @property
    def yaml_path(self) -> str:
        return self._yaml_path

    @property
    def value_pattern(self) -> str:
        return self._value_pattern

class LennyBotAppConfig:

    def __init__(self) -> None:
        self._name = None
        self._source = LennyBotSourceConfig()
        self._actions = []

    @property
    def name(self) -> str:
        return self._name

    @property
    def source(self) -> LennyBotSourceConfig:
        return self._source

    @property
    def actions(self) -> List[LennyBotActionConfig]:
        return self._actions

class LennyBotGithubPr:

    def __init__(self) -> None:
        self._enabled = False
        self._repository = None
        self._branch_prefix = "lennybot-"

    @property
    def enabled(self) -> str:
        return self._enabled

    @property
    def repository(self) -> str:
        return self._repository

    @property
    def branch_prefix(self) -> str:
        return self._branch_prefix

class LennyBotConfig:

    def __init__(self, filename) -> None:
        self._log = None
        with open(filename) as file_ptr:
            self._data = yaml.safe_load(file_ptr)
        self._state_file = None
        self._github_token = None
        self._github_pr = LennyBotGithubPr()
        self._applications = []
        self._parse()
        self._logging()

    def _parse(self):
        self._parse_data(CONFIGURATION_OPTIONS, self._data, self)
        self._parse_env()

    def _logging(self):
        logging.basicConfig(level=logging.DEBUG,format='%(asctime)s [%(levelname)s] %(name)s: %(message)s')
        self._log = logging.getLogger(self.__class__.__name__)
        self._log.debug("Logging was configured")

    def _parse_data(self, schema, data, target):
        for name, property in schema.items():
            if name not in data.keys():
                continue
            config_type = property["type"]
            if config_type in ["object", "list"]:
                self._parse_nested_data(name, property, data, target)
                continue
            attribute_name = property["attribute"]
            property_value = data[name]
            if config_type == "string":
                property_value = str(property_value)
            setattr(target, attribute_name, property_value)

    def _parse_nested_data(self, name, property, data, target):
        config_type = property["type"]
        if "attribute" in property:
            attribute_name = property["attribute"]
            for item in attribute_name.split("."):
                if attribute_name.endswith(item):
                    continue
                target = getattr(target, item)
        if config_type == "object":
            if "attribute" in property:
                target = getattr(target, property["attribute"].split(".")[-1])
            self._parse_data(property["properties"], data[name], target)
            return
        if config_type == "list":
            attribute_name = attribute_name.split(".")[-1]
            for item in data[name]:
                array_target = globals()[property["class"]]()
                self._parse_data(
                    property["properties"], item, array_target)
                getattr(target, attribute_name).append(array_target)
            return

    def _parse_env(self):
        if "LB_GITHUB_TOKEN" in os.environ.keys():
            self._github_token = os.environ["LB_GITHUB_TOKEN"]
        if "LB_STATE_FILE" in os.environ.keys():
            self._state_file = os.environ["LB_STATE_FILE"]
        # TODO make this generic

    @property
    def state_file(self) -> str:
        return self._state_file

    @property
    def applications(self) -> List[LennyBotAppConfig]:
        return self._applications

    @property
    def github_token(self) -> str:
        return self._github_token

    @property
    def github_pr(self) -> LennyBotGithubPr:
        return self._github_pr





class IAction:

    @property
    def application(self) -> str:
        pass

    @property
    def source_version(self) -> str:
        pass

    @property
    def target_version(self) -> str:
        pass

    def run(self):
        pass


"""A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import pkg_resources


def readme():
    with open('README.md') as f:
        return f.read()


def version():
    with open('version.txt') as f:
        return f.read()


def install_requires():
    with open('requirements.txt') as f:
        reqs = pkg_resources.parse_requirements(f.read())
        return [str(ir) for ir in reqs]


setup(
    name='lennybot',
    version=version(),
    author='Simon Schneider',
    author_email='dev@raynigon.com',
    description='Automatic Updates for Kustomize Resources',
    long_description=readme(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent'
    ],
    keywords='kustomize',
    url='http://github.com/raynigon/lennybot',
    project_urls={
        'Bug Reports': 'https://github.com/raynigon/lennybot/issues',
        'Source': 'https://github.com/raynigon/lennybot/',
    },
    install_requires=install_requires(),
    extras_require={
        'dev': ['setuptools', 'wheel'],
        'test': ['coverage'],
    },
    python_requires='>=3.6, <4',
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    data_files=[('generic', ['version.txt', 'README.md', 'requirements.txt'])],
    entry_points={
        'console_scripts': [
            'lennybot=lennybot:main',
        ],
    }
)



import re
from ..config.config import LennyBotActionConfig
from .iaction import IAction


class UpdateDockerfileAction(IAction):

    FROM_PATTERN = r"FROM ([^:]*):[^\s]*(.*)"

    def __init__(self, name, source_version, target_version, config: LennyBotActionConfig) -> None:
        self._name = name
        self._source_version = source_version
        self._target_version = target_version
        self._target_file = config.target_file
        self._image_name = config.image
        if config.value_pattern is not None:
            self._value_pattern = config.value_pattern
        else:
            self._value_pattern = "{{version}}"

    @property
    def application(self) -> str:
        return self._name

    @property
    def source_version(self) -> str:
        return self._source_version

    @property
    def target_version(self) -> str:
        return self._target_version

    def run(self):
        with open(self._target_file, "r") as file_ptr:
            lines = file_ptr.readlines()
        result = []
        for line in lines:
            match = re.match(self.FROM_PATTERN, line)
            if match is not None and match.group(1) == self._image_name:
                result.append(f"FROM {self._image_name}:{self._create_value()}{match.group(2)}\n")
                continue
            result.append(line)
        with open(self._target_file, "w") as file_ptr:
            file_ptr.writelines(result)

    def _create_value(self):
        return self._value_pattern.replace("{{version}}", self._target_version)
